
from dielectric.classes.constants import *  # imports numpy as np
from dielectric.utils.file_utils import load_config

import logging

# logging.basicConfig(level="DEBUG")
logging.basicConfig(level="INFO")
# logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)
# logger.setLevel(logging.INFO)


class Ion:
    def __init__(self, name, mass, radius, D, Hamaker, z, nu):
        self.name = name
        self.mass = mass
        self.radius = radius
        self.D = D
        self.Hamaker = Hamaker
        self.z = z
        self.nu = nu

    @classmethod
    def from_dict(cls, d):
        """ classmethod to enable constructing an instance from configuration file.
        """
        return cls(name=d["name"], mass=d["mass"], radius=d["radius"],
                   D=d["D"], Hamaker=d["Hamaker"],
                   z=d["z"], nu=d["nu"])

    def update(self):
        pass


class Salt:
    def __init__(self):
        self.name = None
        self.conc = None
        self.ions = []
        self.kappa2 = None
        self.D_zero = None  # D_zero of salt: sum over ions

    def calc_kappa2(self):
        """calculation of kappa_zero^2
           calculation of kappa (inverse of Debije length) and kappa^2.

        """
        # calculate kappa2 / kappa: inverse of the Debye length (squared)
        self.kappa2 = 0.0
        for ion in self.ions:
            self.kappa2 += KAPPA2_ZERO * self.conc * ion.nu * np.square(ion.z)

        # logging.info(f" kappa2: {self.kappa2} kappa: '{self.kappa} Debye_length: {self.Debye_length}")

    def calc_D_zero(self):
        nom = 0.0
        denom = 0.0
        for ion in self.ions:
            nom += np.abs(ion.z)*ion.D
            denom += np.abs(ion.z)
        self.D_zero = nom/denom


class Electrolyte2:
    def __init__(self):
        # basic attributes
        self.config_file = None
        self.num_salts = None
        self.salts = []
        # calculated attributes
        self.kappa2 = None
        self.kappa = None
        self.Debye_length = None
        self.D_zero = None         # D_zero of electrolyte: sum over salts
        self.Dc = None
        self.Dn = None
        self.Dt = None             # used for omega << k^2*Dzero
        self.sigma = None          # conductivity: was K1
        self.lambda_c2 = None
        self.lambda_n2 = None

    @classmethod
    def from_yaml(cls, config_yaml: str):
        """ classmethod to enable constructing an instance from configuration file.
        """
        el = cls()
        el.config_file = config_yaml
        d = load_config(el.config_file)
        el.num_salts = len(d["electrolyte"])
        for n in range(el.num_salts):
            ed = d.get("electrolyte")[n]
            salt = Salt()
            salt.name = ed['name']
            salt.conc = ed["conc"]
            num_ions = len(ed["ions"])
            for i in range(num_ions):
                idict = ed.get("ions")[i]
                ion = Ion.from_dict(idict)
                # append by reference, therefore new Ion object in each iteration
                salt.ions.append(ion)
            # append by reference, therefore new Salt object in each iteration
            el.salts.append(salt)
            logging.info(f" salt '{salt.name}' appended to electrolyte")

        el.calc_kappa()
        el.calc_D_zero()
        el.calc_Dc()
        el.calc_Dn_binary()
        el.calc_Dt_binary()
        el.calc_sigma()  # limiting conductivity value
        return el

    def update(self):
        # self.name =
        pass

    def calc_kappa(self):
        """calculation of kappa^2
           calculation of kappa (inverse of Debije length).
        """
        self.kappa2 = 0.0
        self.kappa = 0.0
        self.Debye_length = 0.0
        factor = (np.square(E_CHARGE) * N_AVO) / (EPS_ZW * BOLTZ_T)

        for salt in self.salts:
            for ion in salt.ions:
                self.kappa2 += factor * salt.conc * ion.nu * np.square(ion.z)

        self.kappa = np.sqrt(self.kappa2)  # inverse of the Debije length
        self.Debye_length = np.reciprocal(self.kappa)
        # logging.info(f" kappa2: {self.kappa2} kappa: '{self.kappa} Debye_length: {self.Debye_length}")

    def calc_D_zero(self):
        """ calculates D_zero as sum over ions: |z_i|*D_i over |z_i|.
        """
        if self.salts:
            nominator = np.sum(np.abs(ion.z) * ion.D for salt in self.salts for ion in salt.ions)
            denominator = np.sum(np.abs(ion.z) for salt in self.salts for ion in salt.ions)
            self.D_zero = nominator/denominator

    def calc_sigma(self):
        """calculates limiting conductivity value of electolyte at infinite dilution.
        """
        if self.kappa2 and self.D_zero:
            self.sigma = EPS_ZW * self.kappa2 * self.D_zero  # limiting conductivity value

    def calc_Dc(self):
        """calculates Dc as sum over ions:  |z_i| over |z_i|/D_i/.
        """
        if self.salts:
            denominator = np.sum(np.abs(ion.z) / ion.D for salt in self.salts for ion in salt.ions)
            nominator = np.sum(np.abs(ion.z) for salt in self.salts for ion in salt.ions)
            self.Dc = nominator/denominator

    def calc_Dn_binary(self):
        """calculates Dn as sum over ions:  |z_i| over |z_i|/D_j.ne.i.

        See: https://www.geeksforgeeks.org/python-multiply-all-cross-list-element-pairs/
             https://www.geeksforgeeks.org/python-multiply-numbers-list-3-different-ways/
             https://stackoverflow.com/questions/2853212/all-possible-permutations-of-a-set-of-lists-in-python
             https://numpy.org/doc/stable/reference/generated/numpy.outer.html
             https://numpy.org/doc/stable/reference/generated/numpy.einsum.html
        """
        if self.salts:
            z_list = [np.abs(ion.z) for salt in self.salts for ion in salt.ions]
            one_over_D_list = [1.0/ion.D for salt in self.salts for ion in salt.ions]
            out_mat = np.outer(z_list, one_over_D_list)
            np.fill_diagonal(out_mat, 0.0)
            denominator = np.sum(out_mat)
            nominator = np.sum(z_list)
            self.Dn = nominator/denominator

    def calc_Dt_binary(self):
        """calculates Dt as sum over ions:  |z_i| over |z_i|/D_j.ne.i.

        """
        if self.salts:
            z_list = [ion.z for salt in self.salts for ion in salt.ions]
            z_abs_list = [np.abs(ion.z) for salt in self.salts for ion in salt.ions]
            one_over_D_list = [1.0/ion.D for salt in self.salts for ion in salt.ions]

            dpm = sum(one_over_D_list[::2]) - sum(one_over_D_list[1::2])
            nominator = np.product(z_list) * np.square(dpm)

            out_mat = np.outer(z_abs_list, one_over_D_list)
            np.fill_diagonal(out_mat, 0.0)
            denominator = np.sum(z_abs_list) * np.sum(out_mat)
            self.Dt = nominator/denominator
            self.Dt = np.reciprocal(self.Dt)


if __name__ == "__main__":
    from pathlib import Path
    DATADIR = Path(__file__).parent.parent.parent.parent.absolute().joinpath("ImpedanceData")
    e = Electrolyte(str(DATADIR / "Cell_1_KCl_1mM.yaml"))
    print(f"kappa^2: {e.kappa2} [m^2]")

    IVPDIR = Path(__file__).parent.parent.parent.parent.absolute().joinpath("ImpedanceData")
    e2 = Electrolyte2.from_yaml(str(IVPDIR / "BaCl2.yaml"))
    for salt in e2.salts:
        print(f"{salt.name}")
    print(f"kappa^2: {e2.kappa2} [m^2]")

